import ast
import atexit
import glob
import os
from contextlib import contextmanager
from os.path import expanduser
from typing import List, Iterable

import mlflow
from mlflow.tracking import MlflowClient
from mlflow.utils.autologging_utils import try_mlflow_log

from pypads import logger
from pypads.autolog.hook import Hook
from pypads.autolog.mappings import AlgorithmMapping, MappingRegistry, AlgorithmMeta
from pypads.autolog.pypads_import import extend_import_module, duck_punch_loader
from pypads.caches import PypadsCache, Cache
from pypads.functions.analysis.call_tracker import CallTracker
from pypads.functions.analysis.strace import STrace
from pypads.functions.analysis.validation.parameters import Parameters
from pypads.functions.loggers.base_logger import LoggingFunction
from pypads.functions.loggers.data_flow import Input, Output
from pypads.functions.loggers.debug import LogInit, Log
from pypads.functions.loggers.hardware import Disk, Ram, Cpu
from pypads.functions.loggers.metric import Metric
from pypads.functions.loggers.mlflow.mlflow_autolog import MlflowAutologger
from pypads.functions.loggers.pipeline_detection import PipelineTracker
from pypads.functions.post_run.post_run import PostRunFunction
from pypads.functions.pre_run.git import IGit
from pypads.functions.pre_run.hardware import ISystem, IRam, ICpu, IDisk, IPid, ISocketInfo, IMacAddress
from pypads.functions.pre_run.pre_run import RunInfo, RunLogger, PreRunFunction
from pypads.logging_util import WriteFormats, try_write_artifact, try_read_artifact, get_base_folder
from pypads.util import get_class_that_defined_method, dict_merge

tracking_active = None


class FunctionRegistry:
    """
    This class holds function mappings. Logging functionalities get a name and a underlying function.
    Example.: parameters -> function logging the parameters of the library calls.
    {
    "parameters": <fn>,
    "model": <fn>
    }
    """

    def __init__(self, mapping=None):
        if mapping is None:
            mapping = {}
        self.fns = {}
        for key, value in mapping.items():
            if isinstance(value, Iterable):
                self._add_functions(key, value)
            elif callable(value):
                self._add_functions(key, {value})

    def find_functions(self, name, lib=None, version=None):
        if (name, lib, version) in self.fns:
            return self.fns[(name, lib, version)]
        elif (name, lib) in self.fns:
            return self.fns[(name, lib)]
        elif name in self.fns:
            return self.fns[name]
        else:
            pass
            # logger.warning("Function call with name '" + name + "' is not linked with any logging functionality.")

    def add_functions(self, name, lib=None, version=None, *args: LoggingFunction):
        if lib:
            if version:
                key = (name, lib, version)
            else:
                key = (name, lib)
        else:
            key = name
        self._add_functions(key, args)

    def _add_functions(self, key, fns):
        if key not in self.fns:
            self.fns[key] = set()
        for fn in fns:
            self.fns[key].add(fn)


# --- Pypads App ---
# Default init_run fns
DEFAULT_INIT_RUN_FNS = [RunInfo(), RunLogger(), IGit(_pypads_timeout=3), ISystem(), IRam(), ICpu(), IDisk(), IPid(),
                        ISocketInfo(),
                        IMacAddress(), STrace()]


# Default event mappings. We allow to log parameters, output or input
DEFAULT_LOGGING_FNS = {
    "parameters": Parameters(),
    "output": Output(_pypads_write_format=WriteFormats.text.name),
    "input": Input(_pypads_write_format=WriteFormats.text.name),
    "hardware": {Cpu(), Ram(), Disk()},
    "metric": Metric(),
    "autolog": MlflowAutologger(),
    "pipeline": PipelineTracker(_pypads_pipeline_type="normal", _pypads_pipeline_args=False),
    "log": Log(),
    "init": LogInit()
}

# Default config.
# Pypads mapping files shouldn't interact directly with the logging functions,
# but define events on which different logging functions can listen.
# This config defines such a listening structure.
# {"recursive": track functions recursively. Otherwise check the callstack to only track the top level function.}
DEFAULT_CONFIG = {"events": {
    "init": {"on": ["pypads_init"]},
    "parameters": {"on": ["pypads_fit"]},
    "hardware": {"on": ["pypads_fit"]},
    "output": {"on": ["pypads_fit", "pypads_predict"]},
    "input": {"on": ["pypads_fit"], "with": {"_pypads_write_format": WriteFormats.text.name}},
    "metric": {"on": ["pypads_metric"]},
    "pipeline": {"on": ["pypads_fit", "pypads_predict", "pypads_transform", "pypads_metric"]},
    "log": {"on": ["pypads_log"]}
},
    "track_sub_processes": False,
    "recursion_identity": False,
    "recursion_depth": -1,
    "log_on_failure": True}

# Tag name to save the config to in mlflow context.
CONFIG_NAME = "pypads.config"

"""
TODO keras:
Logs loss and any other metrics specified in the fit
    function, and optimizer data as parameters. Model checkpoints
    are logged as artifacts to a 'models' directory.
    EarlyStopping Integration with Keras Automatic Logging
    MLflow will detect if an ``EarlyStopping`` callback is used in a ``fit()``/``fit_generator()``
    call, and if the ``restore_best_weights`` parameter is set to be ``True``, then MLflow will
    log the metrics associated with the restored model as a final, extra step. The epoch of the
    restored model will also be logged as the metric ``restored_epoch``.
    This allows for easy comparison between the actual metrics of the restored model and
    the metrics of other models.
    If ``restore_best_weights`` is set to be ``False``,
    then MLflow will not log an additional step.
    Regardless of ``restore_best_weights``, MLflow will also log ``stopped_epoch``,
    which indicates the epoch at which training stopped due to early stopping.
    If training does not end due to early stopping, then ``stopped_epoch`` will be logged as ``0``.
    MLflow will also log the parameters of the EarlyStopping callback,
    excluding ``mode`` and ``verbose``.
"""


class PypadsApi:
    def __init__(self, pypads):
        self._pypads = pypads

    # noinspection PyMethodMayBeStatic
    def track(self, fn, ctx=None, events: List = None, mapping: AlgorithmMapping = None):
        if events is None:
            events = ["pypads_log"]
        if ctx is not None and not hasattr(ctx, fn.__name__):
            logger.warning("Given context " + str(ctx) + " doesn't define " + str(fn.__name__))
            ctx = None
        if mapping is None:
            logger.warning("Tracking a function without a mapping definition. A default mapping will be generated.")
            if '__file__' in fn.__globals__:
                lib = fn.__globals__['__file__']
            else:
                lib = fn.__module__
            if ctx is not None:
                if hasattr(ctx, '__module__') and ctx.__module__ is not str.__class__.__module__:
                    ctx_path = ctx.__module__.__name__
                else:
                    ctx_path = ctx.__name__
            else:
                ctx_path = "<unbound>"

            # For all events we want to hook to
            hooks = [Hook(e) for e in events]
            mapping = AlgorithmMapping(ctx_path + "." + fn.__name__, lib, AlgorithmMeta(fn.__name__, []), None,
                                       hooks=hooks)
        return self._pypads.wrap_manager.wrap(fn, ctx=ctx, mapping=mapping)

    def start_run(self, run_id=None, experiment_id=None, run_name=None, nested=False):
        out = mlflow.start_run(run_id=run_id, experiment_id=experiment_id, run_name=run_name, nested=nested)
        self._pypads.run_init_fns()
        return out

    # ---- logging ----
    def log_artifact(self, local_path, meta=None):
        try_mlflow_log(mlflow.log_artifact, local_path)
        self._write_meta(_to_artifact_meta_name(os.path.basename(local_path)), meta)

    def log_mem_artifact(self, name, obj, write_format=WriteFormats.text.name, preserve_folder=True, meta=None):
        try_write_artifact(name, obj, write_format, preserve_folder)
        self._write_meta(_to_artifact_meta_name(name), meta)

    def log_metric(self, key, value, step=None, meta=None):
        mlflow.log_metric(key, value, step)
        self._write_meta(_to_metric_meta_name(key), meta)

    def log_param(self, key, value, meta=None):
        mlflow.log_param(key, value)
        self._write_meta(_to_param_meta_name(key), meta)

    def set_tag(self, key, value):
        return mlflow.set_tag(key, value)

    def _write_meta(self, name, meta):
        if meta:
            try_write_artifact(name + ".meta", meta, WriteFormats.text, preserve_folder=True)

    def _read_meta(self, name):
        # TODO format / json / etc?
        return try_read_artifact(name + ".meta.txt")

    def metric_meta(self, name):
        return self._read_meta(_to_metric_meta_name(name))

    def param_meta(self, name):
        return self._read_meta(_to_param_meta_name(name))

    def artifact_meta(self, name):
        return self._read_meta(_to_artifact_meta_name(name))

    # !--- logging ----

    # ---- run management ----
    @contextmanager
    def intermediate_run(self, **kwargs):
        enclosing_run = mlflow.active_run()
        try:
            run = self._pypads.api.start_run(**kwargs, nested=True)
            self._pypads.cache.run_add("enclosing_run", enclosing_run)
            yield run
        finally:
            if not mlflow.active_run() is enclosing_run:
                self._pypads.api.end_run()
                self._pypads.cache.run_clear()
                self._pypads.cache.run_delete()
                mlflow.start_run(run_id=enclosing_run.info.run_id)

    def _get_pre_run_cache(self):
        if not self._pypads.cache.exists("pre_run_fns"):
            pre_run_fn_cache = Cache()
            self._pypads.cache.run_add("pre_run_fns", pre_run_fn_cache)
        return self._pypads.cache.get("pre_run_fns")

    def register_pre(self, name, pre_fn: PreRunFunction, silent=True):
        cache = self._get_pre_run_cache()
        if cache.exists(name):
            if not silent:
                logger.debug("Pre run fn with name '" + name + "' already exists. Skipped.")
        else:
            cache.add(name, pre_fn)

    def register_pre_fn(self, name, fn, log_function=None, nested=True, intermediate=True, order=0, silent=True):
        self.register_pre(name, PreRunFunction(fn=fn, nested=nested, intermediate=intermediate, order=order,
                                               log_function=log_function), silent=silent)

    def _get_post_run_cache(self):
        if not self._pypads.cache.run_exists("post_run_fns"):
            post_run_fn_cache = Cache()
            self._pypads.cache.run_add("post_run_fns", post_run_fn_cache)
        return self._pypads.cache.run_get("post_run_fns")

    def register_post(self, name, post_fn: PostRunFunction, silent=True):
        cache = self._get_post_run_cache()
        if cache.exists(name):
            if not silent:
                logger.debug("Post run fn with name '" + name + "' already exists. Skipped.")
        else:
            cache.add(name, post_fn)

    def register_post_fn(self, name, fn, log_function=None, nested=True, intermediate=True, order=0, silent=True):
        self.register_post(name, post_fn=PostRunFunction(fn=fn, nested=nested, intermediate=intermediate, order=order,
                                                         log_function=log_function), silent=silent)

    def active_run(self):
        return mlflow.active_run()

    def is_intermediate_run(self):
        enclosing_run = self._pypads.cache.run_get("enclosing_run")
        return enclosing_run is not None

    def end_run(self):
        run = self.active_run()

        chached_fns = self._get_post_run_cache()
        fn_list = [v for i, v in chached_fns.items()]
        fn_list.sort(key=lambda t: t.order())
        for fn in fn_list:
            fn(self, _pypads_env=None)

        mlflow.end_run()

        # --- Clean tmp files in disk cache after run ---
        folder = get_base_folder(run)
        if os.path.exists(folder):
            import shutil
            shutil.rmtree(folder)
        # !-- Clean tmp files in disk cache after run ---
    # !--- run management ----


def _to_artifact_meta_name(name):
    return name + ".artifact"


def _to_metric_meta_name(name):
    return name + ".metric"


def _to_param_meta_name(name):
    return name + ".param"


class PypadsDecorators:
    def __init__(self, pypads):
        self._pypads = pypads

    def track(self, event="pypads_log", mapping: AlgorithmMapping = None):
        def track_decorator(fn):
            ctx = get_class_that_defined_method(fn)
            events = event if isinstance(event, List) else [event]
            return self._pypads.api.track(ctx=ctx, fn=fn, events=events, mapping=mapping)

        return track_decorator


# TODO pypads isn't allowed to hold a state anymore (Everything with state should be part of the caching system)
#  - We want to be able to rebuild PyPads from the config and cache alone if possible
#  to stop the need for pickeling pypads as a whole.

class PyPads:
    """
    PyPads app and base class. It enable automatic logging for all libraries included in the mapping files.
    Serves as the main entrypoint to PyPads. After constructing this app tracking is activated.

        :param uri: **string, optional (default=None)** Address of local or remote tracking server that **MLflow** uses to record runs. If None, then it tries to get the environment variable **'MLFLOW_PATH'** or the **'HOMEPATH'** of the user.
        :param name: **string, optional (default=None)** Name of the **MLflow** experiment to track.
        :param mapping_paths: **list, optional (default=None)** Absolute paths to additional mapping files.
        :param mapping: **dict, optional (default=None)** Mapping to the logging functions to use for the tracking of the events. If None, then a DEFAULT_MAPPING is used which allow to log parameters, outputs or inputs.
        :param init_run_fns: **list, optional (default=None)** Logging function to execute on tracking initialization.
        :param include_default_mappings: **boolean, optional (default=True)** A flag whether to use the default provided mappings or not.
        :param logging_fns: **dict, optional (default=None)** User defined logging functions to use where each dict item has to be ' "event": fn' or ' "event": {fn1,fn2,...}'.
        :param config: **dict, optional (default=None)** A dictionary that maps the events defined in PyPads mapping files with the logging functions.
        :param reload_modules: **boolean, optional (default=False)** Reload and duck punch already loaded modules before the tracking activation if set to True.
        :param clear_imports: **boolean, optional (default=False)** Delete alredy loaded modules for sys.modules() if set to True.

    """

    def __init__(self, uri=None, name=None, mapping_paths=None, mapping=None, init_run_fns=None,
                 include_default_mappings=True,
                 logging_fns=None, config=None, reload_modules=False, reload_warnings=True, clear_imports=False,
                 affected_modules=None, pre_initialized_cache=None, disable_run_init=True):
        """
        :param uri: **string, optional (default=None)** <br> Address of local or remote tracking server that **MLflow** uses to record runs. If None, then it tries to get the environment variable **'MLFLOW_PATH'** or the **'HOMEPATH'** of the user.
        :param name: **string, optional (default=None)** <br> Name of the **MLflow** experiment to track.
        :param mapping_paths: **list, optional (default=None)** <br> Absolute paths to additional mapping files.
        :param mapping: **dict, optional (default=None)** <br> Mapping to the logging functions to use for the tracking of the events. If None, then a DEFAULT_MAPPING is used which allow to log parameters, outputs or inputs.
        :param init_run_fns: **list, optional (default=None)** <br> Logging function to execute on tracking initialization.
        :param include_default_mappings: **boolean, optional (default=True)** <br> A flag whether to use the default provided mappings or not.
        :param logging_fns: **dict, optional (default=None)** <br> User defined logging functions to use where each dict item has to be ' "event": fn' or ' "event": {fn1,fn2,...}'.
        :param config: **dict, optional (default=None)** <br> A dictionary that maps the events defined in PyPads mapping files with the logging functions.
        :param reload_modules: **boolean, optional (default=False)** <br> Reload and duck punch already loaded modules before the tracking activation if set to True.
        :param clear_imports: **boolean, optional (default=False)** <br> Delete alredy loaded modules for sys.modules() if set to True.
        """
        from pypads.pypads import set_current_pads
        set_current_pads(self)

        # Init variable to filled later in this constructor
        self._config = None
        self._atexit_fns = []

        if mapping_paths is None:
            mapping_paths = []

        if init_run_fns is None:
            init_run_fns = DEFAULT_INIT_RUN_FNS

        from pypads.autolog.wrapping.wrapping import WrapManager
        self._wrap_manager = WrapManager(self)
        self._api = PypadsApi(self)
        self._decorators = PypadsDecorators(self)

        self._cache = pre_initialized_cache if pre_initialized_cache else PypadsCache()
        self._call_tracker = CallTracker(self)
        self._init_mapping_registry(*mapping_paths, mapping=mapping, include_defaults=include_default_mappings)

        def cleanup():
            from pypads.pypads import get_current_pads
            pads: PyPads = get_current_pads()
            if pads.api.active_run():
                pads.api.end_run()

        self.add_atexit_fn(cleanup)

        self._init_run_fns = init_run_fns
        self._init_mlflow_backend(uri, name, config, disable_run_init)
        self._function_registry = FunctionRegistry(logging_fns or DEFAULT_LOGGING_FNS)
        self.activate_tracking(reload_modules=reload_modules, reload_warnings=reload_warnings,
                               clear_imports=clear_imports, affected_modules=affected_modules)

    def add_atexit_fn(self, fn):
        """
        Add function to be executed before stopping your process.
        """
        self._atexit_fns.append(fn)
        atexit.register(fn)

    def _is_affected_module(self, name, affected_modules=None):
        if affected_modules is None:
            affected_modules = self.wrap_manager.module_wrapper.punched_module_names

        affected = set([module.split(".", 1)[0] for module in affected_modules])
        return any([name.startswith(module + ".") or name == module for module in affected]) or any(
            [name.startswith(module + ".") or name == module for module in
             affected_modules])

    def activate_tracking(self, reload_modules=False, reload_warnings=True, clear_imports=False, affected_modules=None):
        """
        Function to duck punch all objects defined in the mapping files. This should at best be called before importing
        any libraries.
        :param mod_globals: globals() object used to duckpunch already loaded classes
        :return:
        """
        if affected_modules is None:
            affected_modules = self.wrap_manager.module_wrapper.punched_module_names | self.mapping_registry.get_libraries()

        global tracking_active
        if not tracking_active:
            from pypads.pypads import set_current_pads
            set_current_pads(self)

            # Add our loader to the meta_path
            extend_import_module()

            import sys
            import importlib
            loaded_modules = [(name, module) for name, module in sys.modules.items()]
            for name, module in loaded_modules:
                if self._is_affected_module(name, affected_modules):
                    if reload_warnings:
                        logger.warning(
                            name + " was imported before PyPads. To enable tracking import PyPads before or use "
                                   "reload_modules. Every already created instance is not tracked.")

                    if clear_imports:
                        del sys.modules[name]

                    if reload_modules:
                        try:
                            spec = importlib.util.find_spec(module.__name__)
                            duck_punch_loader(spec)
                            loader = spec.loader
                            module = loader.load_module(module.__name__)
                            loader.exec_module(module)
                            importlib.reload(module)
                        except Exception as e:
                            logger.debug("Couldn't reload module " + str(e))

            tracking_active = True
        else:
            raise Exception("Currently only one tracker can be activated at once.")

    def deactivate_tracking(self, run_atexits=False, reload_modules=True):
        # run atexit fns if needed
        if run_atexits:
            for fn in self._atexit_fns:
                fn()

        # Remove atexit fns
        for fn in self._atexit_fns:
            atexit.unregister(fn)

        import sys
        import importlib
        loaded_modules = [(name, module) for name, module in sys.modules.items()]
        for name, module in loaded_modules:
            if self._is_affected_module(name):
                del sys.modules[name]

                if reload_modules:
                    # reload modules if they where affected
                    try:
                        spec = importlib.util.find_spec(module.__name__)
                        duck_punch_loader(spec)
                        loader = spec.loader
                        module = loader.load_module(module.__name__)
                        loader.exec_module(module)
                        importlib.reload(module)
                    except Exception as e:
                        logger.debug("Couldn't reload module " + str(e))

        global tracking_active
        tracking_active = False
        # noinspection PyTypeChecker
        from pypads.pypads import set_current_pads
        set_current_pads(None)

    def _init_mlflow_backend(self, uri=None, name=None, config=None, disable_run_init=True):
        """
        Intialize the mlflow backend experiment and run as well as store the config to it.
        :param uri:
        :param name:
        :param config:
        :return:
        """
        self._uri = uri or os.environ.get('MLFLOW_PATH') or 'file:' + os.path.expanduser('~/.mlruns')
        mlflow.set_tracking_uri(self._uri)

        # check if there is already an active run
        run = mlflow.active_run()
        if run is None:
            # Create run if run doesn't already exist
            name = name or "Default-PyPads"
            experiment = mlflow.get_experiment_by_name(name)
            experiment_id = experiment.experiment_id if experiment else mlflow.create_experiment(name)
            run = self.api.start_run(experiment_id=experiment_id)
        else:
            # Run init functions if run already exists but tracking is starting for it now
            if not disable_run_init:
                self.run_init_fns()
        _experiment = self.mlf.get_experiment_by_name(name) if name else self.mlf.get_experiment(
            run.info.experiment_id)
        if config:
            self.config = dict_merge({"events": {}}, DEFAULT_CONFIG, config)
        else:
            self.config = dict_merge({"events": {}}, DEFAULT_CONFIG)

        # override active run if used
        if name and run.info.experiment_id is not _experiment.experiment_id:
            logger.warning("Active run doesn't match given input name " + name + ". Recreating new run.")
            try:
                self.api.start_run(experiment_id=_experiment.experiment_id)
            except Exception:
                mlflow.end_run()
                self.api.start_run(experiment_id=_experiment.experiment_id)

    def _init_mapping_registry(self, *paths, mapping=None, include_defaults=True):
        """
        Function to initialize the mapping file registry
        :param paths:
        :param mapping:
        :param include_defaults:
        :return:
        """
        mapping_file_paths = []
        if include_defaults:
            # Use our with the package delivered mapping files
            mapping_file_paths.extend(glob.glob(os.path.join(expanduser("~"), ".pypads", "bindings", "**.json")))
            mapping_file_paths.extend(glob.glob(
                os.path.abspath(
                    os.path.join(os.path.dirname(__file__), "bindings", "resources", "mapping", "**.json"))))
        if paths:
            mapping_file_paths.extend(paths)
        self._mapping_registry = MappingRegistry(*mapping_file_paths)
        if mapping:
            if isinstance(mapping, dict):
                for key, mapping in mapping.items():
                    self._mapping_registry.add_mapping(mapping, key=key)
            else:
                self._mapping_registry.add_mapping(mapping, key=id(mapping))

    @property
    def tracking_uri(self):
        return self._uri

    @property
    def wrap_manager(self):
        return self._wrap_manager

    @property
    def mapping_registry(self):
        return self._mapping_registry

    @property
    def mlf(self) -> MlflowClient:
        return MlflowClient(self._uri)

    @property
    def function_registry(self) -> FunctionRegistry:
        return self._function_registry

    @property
    def config(self):
        if self._config:
            return self._config
        tags = self.mlf.get_run(mlflow.active_run().info.run_id).data.tags
        if CONFIG_NAME not in tags:
            raise Exception("Config for pypads is not defined.")
        try:
            return ast.literal_eval(tags[CONFIG_NAME])
        except Exception as e:
            raise Exception("Config for pypads is malformed. " + str(e))

    @config.setter
    def config(self, value: dict):
        mlflow.set_tag(CONFIG_NAME, value)
        self._config = value

    @property
    def api(self) -> PypadsApi:
        return self._api

    @property
    def decorators(self):
        return self._decorators

    @property
    def call_tracker(self):
        return self._call_tracker

    @property
    def cache(self):
        return self._cache

    def run_init_fns(self):
        self._init_run_fns.sort(key=lambda f: f.order())
        for fn in self._init_run_fns:
            if callable(fn):
                fn(self, _pypads_env=None)



import re
from _py_abc import ABCMeta
from abc import abstractmethod

from pypads import logger


class Hook:
    """
    This class defines a pypads hook. This hook always triggers. Hooks are injected into function calls to inject different functionality.
    """

    type = "always"

    def __init__(self, event):
        self._event = event

    @property
    def event(self):
        """
        Event to trigger on hook execution
        :return:
        """
        return self._event

    @classmethod
    def has_type_name(cls, type):
        return cls.type == type

    def is_applicable(self, *args, **kwargs):
        return True

    def __eq__(self, other):
        return self.event == other.event and type(self) == type(other)

    def __hash__(self):
        return hash(self.event)


class RegexHook(Hook):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, event, regex):
        super().__init__(event)
        try:
            self._regex = re.compile(regex)
        except Exception as e:
            logger.error("Couldn't compile regex: " + str(regex) + "of hook" + str(self) + ". Disabling it.")
            # Regex to never match anything
            self._regex = re.compile('a^')

    @property
    def regex(self):
        """
        Function name to hook to
        :return:
        """
        return self._regex

    def __eq__(self, other):
        return super().__eq__(other) and self.regex == other.regex

    def __hash__(self):
        return hash(self.event) + hash(self.regex) + hash(type(self))


class QualNameHook(RegexHook):
    """
    This class defines a pypads hook triggering if the function name equals the given name.
    """

    def __init__(self, event, regex):
        super().__init__(event, regex)

    type = "qual_name"

    def is_applicable(self, *args, fn=None, **kwargs):
        return fn is not None and hasattr(fn, "__name__") and self.regex.match(fn.__name__)


class PackageNameHook(RegexHook):
    """
    This class defines a pypads hook triggering if the package name includes given name.
    """

    def __init__(self, event, regex):
        super().__init__(event, regex)

    type = "package_name"

    def is_applicable(self, *args, mapping=None, **kwargs):
        return mapping is not None and self.regex.match(mapping.reference)


def get_hooks(hook_map):
    """
    This function returns hook objects defined in a mapping.
    :param hook_map: mapping containing the hooks
    :return:
    """
    hooks = []
    for event, hook_serialization in hook_map.items():
        if Hook.has_type_name(hook_serialization):
            # If we are a string "always"
            hooks.append(Hook(event))
        else:
            for hook in hook_serialization:
                # If hook is already a hook
                if isinstance(hook, Hook):
                    hooks.append(hook)

                # If hook dict has a type
                elif hasattr(hook, 'type'):
                    if QualNameHook.has_type_name(hook['type']):
                        hooks.append(QualNameHook(event, hook['regex']))
                    elif PackageNameHook.has_type_name(hook['type']):
                        hooks.append(PackageNameHook(event, hook['regex']))
                    else:
                        logger.warning("Type " + str(hook['type']) + " of hook " + str(hook) + " unknown.")

                # If hook is just a string
                else:
                    hooks.append(QualNameHook(event, hook))
    return hooks


def make_hook_applicable_filter(hook, ctx, mapping):
    """
    Create a filter to check if hook is applicable
    :param hook:
    :param ctx:
    :param mapping:
    :return:
    """

    def hook_applicable_filter(name):
        if hasattr(ctx, name):
            try:
                fn = getattr(ctx, name)
                return hook.is_applicable(mapping=mapping, fn=fn)
            except RecursionError as re:
                logger.error("Recursion error on '" + str(
                    ctx) + "'. This might be because __get_attr__ is being wrapped. " + str(re))
        else:
            logger.warning("Can't access attribute '" + str(name) + "' on '" + str(ctx) + "'. Skipping.")
        return False

    return hook_applicable_filter


def find_applicable_hooks(context, mapping):
    if mapping.hooks:
        for hook in mapping.hooks:
            for name in list(filter(make_hook_applicable_filter(hook, context, mapping), dir(context))):
                yield name, context, mapping



"""
The configuration-file 'mappings.json' is read during import of the module and stored statically inside the module.
"""

import importlib
import json
import os

import pip._internal.utils.misc as pip

type_mappings = {}
name_mappings = {}
alternate_name_mappings = {}
version_mappings = {}

path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../../bindings/resources/mapping"))
mapping_files = os.listdir(path)

# For every file verify whether it is a json file and then add to the current type_mappings and name_mappings
for file in mapping_files:
    if os.path.isfile(os.path.join(path,file)) and file.endswith(".json"):
        try:
            with open(os.path.join(path,file), encoding='utf-8-sig') as f:
                data = json.loads(f.read())
                algorithms = data['algorithms']
                metadata = data['metadata']
                version_mappings[metadata['library']] = metadata

            for alg in algorithms:
                name_mappings[alg['name']] = alg

                for k in alg['implementation']:
                    type_mappings[alg['implementation'][k]] = (alg, k)

                for alt_name in alg['other_names']:
                    # Convert the alternate names to lower case for better usability
                    alternate_name_mappings[str(alt_name).lower()] = str(alg['name']).lower()

        except ValueError:
            raise ValueError('Error when parsing JSON file {name}'.format(name=file))

# Get all the packages
packages = pip.get_installed_distributions()

# Iterate through packages
for package in packages:
    if str(package).lower().startswith('pypa-'):
        package_name = str(package).lower().split(sep=' ')[0].replace('-', '_')
        module = importlib.import_module(package_name)
        if hasattr(module, 'get_mappings'):
            data = module.get_mappings()
            algorithms = data['algorithms']
            metadata = data['metadata']
            version_mappings[metadata['library']] = metadata
            for alg in algorithms:
                name_mappings[alg['name']] = alg

                for k in alg['implementation']:
                    type_mappings[alg['implementation'][k]] = (alg, k)

                for alt_name in alg['other_names']:
                    alternate_name_mappings[alt_name] = alg['name']


supported_frameworks = list(version_mappings.keys())



from pypads import logger

from pypads.autolog.hook import find_applicable_hooks
from pypads.autolog.wrapping.base_wrapper import BaseWrapper


class ClassWrapper(BaseWrapper):

    def __init__(self, pypads):
        super().__init__(pypads)
        self._punched_class_names = set()

    @property
    def punched_class_names(self):
        return self._punched_class_names

    def wrap(self, clazz, context, mapping):
        """
            Wrap a class in given ctx with pypads functionality
            :param clazz:
            :param context:
            :param mapping:
            :return:
            """
        if clazz.__name__ not in self.punched_class_names or not context.has_wrap_meta(mapping, clazz):
            try:
                context.store_wrap_meta(mapping, clazz)
            except Exception:
                return clazz
            self.punched_class_names.add(clazz.__name__)

            if not context.has_original(clazz):
                context.store_original(clazz)

            # Module was changed and should be added to the list of modules which have been changed
            if hasattr(clazz, "__module__"):
                self._pypads.wrap_manager.module_wrapper.punched_module_names.add(clazz.__name__)

            # Get default hooks
            if not mapping.hooks:
                mapping.hooks = mapping.in_collection.get_default_class_hooks()

            # Try to wrap every attr of the class
            for name, c, mapping in find_applicable_hooks(clazz, mapping):
                self._pypads.wrap_manager.wrap(getattr(clazz, name), c, mapping)

            # Override class on module
            reference_name = mapping.reference.rsplit('.', 1)[-1]
            context.overwrite(reference_name, clazz)
        else:
            logger.debug("Class " + str(clazz) + "already duck-puched.")
        return clazz



import inspect
import sys
import types
from functools import wraps
from importlib._bootstrap_external import PathFinder
# noinspection PyUnresolvedReferences
from multiprocessing import Value

from pypads import logger
from pypads.autolog.mappings import AlgorithmMapping
from pypads.autolog.wrapping.base_wrapper import Context


def _add_found_class(mapping):
    from pypads.pypads import get_current_pads
    return get_current_pads().mapping_registry.add_found_class(mapping)


def _get_algorithm_mappings():
    from pypads.pypads import get_current_pads
    return get_current_pads().mapping_registry.get_relevant_mappings()


def _add_inherited_mapping(clazz, super_class):
    from pypads.pypads import get_current_pads
    if clazz.__name__ not in get_current_pads().wrap_manager.class_wrapper.punched_class_names:
        if hasattr(super_class, "_pypads_mapping_" + super_class.__name__):
            for mapping in getattr(super_class, "_pypads_mapping_" + super_class.__name__):
                found_mapping = AlgorithmMapping(
                    clazz.__module__ + "." + clazz.__qualname__, mapping.library,
                    mapping.algorithm, mapping.file, mapping.hooks)
                found_mapping.in_collection = mapping.in_collection
                _add_found_class(mapping=found_mapping)


def duck_punch_loader(spec):
    if not spec.loader:
        return spec

    original_exec = spec.loader.exec_module

    @wraps(original_exec)
    def exec_module(self, module, execute=original_exec):
        out = execute(module)

        # History to check if a class inherits a wrapping intra-module
        mro_entry_history = {}

        from pypads.pypads import current_pads
        if current_pads:
            # TODO we might want to make this configurable/improve performance.
            #  This looks at every imported class and every mapping.
            # On execution of a module we search for relevant mappings
            # For every var on module
            for name in dir(module):
                reference = getattr(module, name)
                if inspect.isclass(reference) and hasattr(reference, "mro"):
                    try:

                        # Look at the MRO and add classes to be punched which inherit from our punched classes
                        mro_ = reference.mro()[1:]
                        for entry in mro_:
                            if entry not in mro_entry_history.keys():
                                mro_entry_history[entry] = [reference]
                            else:
                                mro_entry_history[entry].append(reference)
                        overlap = set(
                            [c.__name__ for c in mro_]) & current_pads.wrap_manager.class_wrapper.punched_class_names
                        if bool(overlap):
                            # TODO Is one inherited mapping maybe enough? Only add for the first one?
                            for o in overlap:
                                _add_inherited_mapping(reference, o)
                    except Exception as e:
                        logger.debug("Skipping superclasses of " + str(reference) + ". " + str(e))

            # For every mapping in our mappings
            for mapping in _get_algorithm_mappings():
                if mapping.reference.startswith(module.__name__):
                    if mapping.reference == module.__name__:
                        current_pads.wrap_manager.wrap(module, None, mapping)
                    else:
                        ref = mapping.reference
                        path = ref[len(module.__name__) + 1:].rsplit(".")
                        obj = module
                        ctx = obj
                        for seg in path:
                            try:
                                ctx = obj
                                obj = getattr(obj, seg)
                            except AttributeError:
                                obj = None
                                break

                        if obj:
                            if inspect.isclass(obj):
                                if mapping.reference == obj.__module__ + "." + obj.__name__:
                                    current_pads.wrap_manager.wrap(obj, Context(ctx), mapping)
                                    if obj in mro_entry_history:
                                        for clazz in mro_entry_history[obj]:
                                            _add_inherited_mapping(clazz, obj)

                            elif inspect.isfunction(obj):
                                current_pads.wrap_manager.wrap(obj, Context(ctx), mapping)
        return out

    spec.loader.exec_module = types.MethodType(exec_module, spec.loader)
    return spec


class PyPadsFinder(PathFinder):
    """
    Import lib extension. This finder provides a special loader if mapping files contain the object.
    """

    def find_spec(cls, fullname, path=None, target=None):
        if fullname not in sys.modules:
            path_ = sys.meta_path[
                    [i for i in range(len(sys.meta_path)) if isinstance(sys.meta_path[i], PyPadsFinder)].pop() + 1:]
            i = iter(path_)
            spec = None
            try:
                # Find a valid importer for the given fullname on the meta_path
                importer = None
                while not spec:
                    importer = next(i)
                    if hasattr(importer, "find_spec"):
                        spec = importer.find_spec(fullname, path)

                # Use the importer as a real importer but wrap it in the PyPadsLoader
                if spec and importer and spec.loader:
                    return duck_punch_loader(spec)
            except StopIteration:
                pass


def extend_import_module():
    """
    Function to add the custom import logic to the python importlib execution
    :return:
    """
    path_finder = [i for i in range(len(sys.meta_path)) if
                   "_frozen_importlib_external.PathFinder" in str(sys.meta_path[i])]
    sys.meta_path.insert(path_finder.pop(), PyPadsFinder())



import inspect


def get_class_that_defined_method(meth):
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def dict_merge(*dicts):
    merged = {}
    for d in dicts:
        if isinstance(d, dict):
            for key, value in d.items():
                if isinstance(value, dict):
                    node = merged.setdefault(key, {})
                    merged[key] = dict_merge(node, value)
                else:
                    merged[key] = value
    return merged


def sizeof_fmt(num, suffix='B'):
    if num == 0:
        return '0'
    import math
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)
    if magnitude > 7:
        return '{:.1f}{}{}'.format(val, 'Yi', suffix)
    return '{:3.1f}{}{}'.format(val, ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z'][magnitude], suffix)


def local_uri_to_path(uri):
    """
    Convert URI to local filesystem path.
    """
    from six.moves import urllib
    path = urllib.parse.urlparse(uri).path if uri.startswith("file:") else uri
    return urllib.request.url2pathname(path)


def is_package_available(name):
    import importlib
    spam_loader = importlib.util.find_spec(name)
    return spam_loader is not None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypads',
 'pypads.autolog',
 'pypads.autolog.wrapping',
 'pypads.bindings',
 'pypads.bindings.resources',
 'pypads.bindings.resources.mapping',
 'pypads.functions',
 'pypads.functions.analysis',
 'pypads.functions.analysis.validation',
 'pypads.functions.analysis.validation.visitors',
 'pypads.functions.loggers',
 'pypads.functions.loggers.mlflow',
 'pypads.functions.post_run',
 'pypads.functions.pre_run',
 'pypads.parallel']

package_data = \
{'': ['*']}

install_requires = \
['boltons>=19.3.0,<20.0.0',
 'cloudpickle>=1.3.0,<2.0.0',
 'loguru>=0.4.1,<0.5.0',
 'mlflow>=1.6.0,<2.0.0']

setup_kwargs = {
    'name': 'pypads',
    'version': '0.1.12',
    'description': 'PyPaDS aims to to add tracking functionality to machine learning libraries.',
    'long_description': '\n# PyPads\nBuilding on the [MLFlow](https://github.com/mlflow/mlflow/) toolset this project aims to extend the functionality for MLFlow, increase the automation and therefore reduce the workload for the user. The production of structured results is an additional goal of the extension.\n\n[![Documentation Status](https://readthedocs.org/projects/pypads/badge/?version=latest)](https://pypads.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/pypads.svg)](https://badge.fury.io/py/pypads)\n\n\n# Intalling\nThis tool requires those libraries to work:\n\n    Python (>= 3.6),\n    cloudpickle (>= 1.3.3),\n    mlflow (>= 1.6.0),\n    boltons (>= 19.3.0),\n    loguru (>=0.4.1)\n    \nPyPads only support python 3.6 and higher. To install pypads run this in you terminal\n\n**Using source code**\n\nFirst, you have to install **poetry**\n\n    poetry build & pip install ./dist/pypads-0.1.0.tar.gz \n \n**Using pip ([PyPi release](https://pypi.org/project/pypads/))**\n\nThe package can be found on PyPi in following [project](https://pypi.org/project/pypads/).\n\n    pip install pypads\n\n### Tests\nThe unit tests can be found under \'test/\' and can be executed using\n\n    poetry run pytest test/\n\n# Documentation\n\nFor more information, look into the [official documentation of PyPads](https://pypads.readthedocs.io/en/latest/).\n\n# Getting Started\n         \n### Usage example\npypads is easy to use. Just define what is needed to be tracked in the config and call PyPads.\n\nA simple example looks like the following,\n```python\nfrom pypads.base import PyPads\n# define the configuration, in this case we want to track the parameters, \n# outputs and the inputs of each called function included in the hooks (pypads_fit, pypads_predict)\nconfig = {"events": {\n    "parameters": {"on": ["pypads_fit"]},\n    "output": {"on": ["pypads_fit", "pypads_predict"]},\n    "input": {"on": ["pypads_fit"]}\n}}\n# A simple initialization of the class will activate the tracking\nPyPads(config=config)\n\n# An example\nfrom sklearn import datasets, metrics\nfrom sklearn.tree import DecisionTreeClassifier\n\n# load the iris datasets\ndataset = datasets.load_iris()\n\n# fit a model to the data\nmodel = DecisionTreeClassifier()\nmodel.fit(dataset.data, dataset.target) # pypads will track the parameters, output, and input of the model fit function.\n# get the predictions\npredicted = model.predict(dataset.data) # pypads will track only the output of the model predict function.\n```\n        \n        \nThe used hooks for each event are defined in the mapping json file where each hook represents the functions to listen to.\nIn the [sklearn mapping](pypads/bindings/resources/mapping/sklearn_0_19_1.json) json file, an example entry would be:\n\n    {\n      "name": "base sklearn estimator",\n      "other_names": [],\n      "implementation": {\n        "sklearn": "sklearn.base.BaseEstimator"\n      },\n      "hooks": {\n        "pypads_fit": [\n          "fit",\n          "fit_predict",\n          "fit_transform"\n        ],\n        "pypads_predict": [\n          "fit_predict",\n          "predict"\n        ],\n        "pypads_transform": [\n          "fit_transform",\n          "transform"\n        ]\n      }\n    }\n\nFor instance, "pypads_fit" is an event listener on any fit, fit_predict and fit_transform call made by the tracked model class which is in this case **BaseEstimator** that most estimators inherits from.\n\n### Defining hooks\nHooks are what triggers an "event" which is associated to one or more logger function in the mapping.\n\n#### Always\n    {\n      "name": "sklearn classification metrics",\n      "other_names": [],\n      "implementation": {\n        "sklearn": "sklearn.metrics.classification"\n      },\n      "hooks": {\n        "pypads_metric": "always"\n      }\n    }\nThis hook triggers always. If you annotate a module with this hook, all its functions and classes will be tracked.\n\n#### QualNameHook\n    {\n      "name": "sklearn classification metrics",\n      "other_names": [],\n      "implementation": {\n        "sklearn": "sklearn.metrics.classification"\n      },\n      "hooks": {\n        "pypads_metric": ["f1_score"]\n      }\n    }\nTracks function with a name matching the given expression by compiling a regex expression.\n\n#### PackageNameHook\n    {\n      "name": "sklearn classification metrics",\n      "other_names": [],\n      "implementation": {\n        "sklearn": "sklearn.metrics"\n      },\n      "hooks": {\n        "pypads_metric": [{"type": "package_name", "name":".*classification.*"}]\n      }\n    }\nTracks all attribute on module where package name is matching Regex.\n\n### Default Events, Hooks and Logging Functions\n\nThe default configuration of events/hooks and logging functions for PyPads:\n\n    DEFAULT_CONFIG = {"events": {\n    "init": {"on": ["pypads_init"]},\n    "parameters": {"on": ["pypads_fit"]},\n    "hardware": {"on": ["pypads_fit"]},\n    "output": {"on": ["pypads_fit", "pypads_predict"]},\n    "input": {"on": ["pypads_fit"], "with": {"_pypads_write_format": WriteFormats.text.name}},\n    "metric": {"on": ["pypads_metric"]},\n    "pipeline": {"on": ["pypads_fit", "pypads_predict", "pypads_transform", "pypads_metric"]},\n    "log": {"on": ["pypads_log"]}\n    },\n        "recursion_identity": False,\n        "recursion_depth": -1,\n        "log_on_failure": True}\n    \n    DEFAULT_LOGGING_FNS = {\n    "parameters": Parameters(),\n    "output": Output(_pypads_write_format=WriteFormats.text.name),\n    "input": Input(_pypads_write_format=WriteFormats.text.name),\n    "hardware": {Cpu(), Ram(), Disk()},\n    "metric": Metric(),\n    "autolog": MlflowAutologger(),\n    "pipeline": PipelineTracker(_pypads_pipeline_type="normal", _pypads_pipeline_args=False),\n    "log": Log(),\n    "init": LogInit()\n\nLoggers in pypads goes into three categories. Pre, Post run loggers and event based loggers.\n\n* Event based loggers:\n\n| Logger  | Event | Hook | Description\n| :-------------: |:----------:|: -----------:| ----------------|\n| LogInit  | init | \'pypads_init\'| Debugging purposes |\n| Log  | log | \'pypads_log\'| Debugging purposes |\n| Parameters  |  parameters | \'pypads_fit\'| tracks parameters of the tracked function call |\n| Cpu,Ram,Disk  |  hardware | \'pypads_fit\'| track usage information, properties and other info on CPU, Memory and Disk. |\n| Input  |  input | \'pypads_fit\' |tracks the input parameters of the current tracked function call. | \n| Output  | output | \'pypads_predict\', \'pypads_fit\' |Logs the output of the current tracked function call.| \n| Metric  | metric | \'pypads_metric\' |tracks the output of the tracked metric function. | \n| PipelineTracker  | pipeline | \'pypads_fit\',\'pypads_predict\', \'pypads_transform\', \'pypads_metrics\'|tracks the workflow of execution of the different pipeline elements of the experiment.| \n\n* Pre/Post run loggers:\n\n| Logger  | Pre/Post | Description\n| :-------------:|: -----------:| ----------------|\n| IGit  | Pre | Source code management and tracking|\n| ISystem  | Pre | System information (os,version,machine...)|\n| ICpu  |  Pre | Cpu information (Nbr of cores, max/min frequency)|\n| IRam  |  Pre | Memory information (Total RAM, SWAP)|\n| IDisk  |  Pre | Disk information (disk total space)| \n| IPid  | Pre | Process information (ID, command, cpu usage, memory usage)| \n| ISocketInfo  | Pre | Network information (hostname, ip address)| \n| IMacAddress  | Pre | Mac address |\n\n# Concept\nLogging results of machine learning workflows often shares similar structures and goals. You might want to track parameteres, loss functions, metrics or other characteristic numbers. While the produced output shares a lot of concepts and could be standardized, implementations are diverse and integrating them or their autologging functionality into such a standard needs manual labour. Each and every version of a library might change internal structures and hard coding interfaces can need intesive work. Pypads aims to feature following main techniques to handle autologging and standardization efforts:\n- **Automatic metric tracking:** TODO\n- **Automatic execution tracking:** TODO \n- **Community driven mapping files:** A means to log data from python libaries like sklearn. Interfaces are not added directly to MLFlows, but derived from versioned mapping files of frameworks.\n- **Output standardization:** TODO\n\n\n### PyPads class\nAs we have seen, a simple initialization of the class at the top of your code activate the tracking for libraries that has a mapping file defining the algorithms to track.\n\nBeside the configuration, **PyPads** takes other optional arguments.\n```python        \nclass PyPads(uri=None, name=None, mapping_paths=None, mapping=None, init_run_fns=None,\n                 include_default_mappings=True,\n                 logging_fns=None, config=None, reload_modules=False, reload_warnings=True, clear_imports=False,\n                 affected_modules=None)\n```\n[Source](https://github.com/padre-lab-eu/pypads/blob/0cb9f9bd5dff7753f7c47dc691d41edd0426a90a/pypads/base.py#L141)\n\n**Parameters**:\n> **uri : string, optional (default=None)** <br> Address of local or remote tracking server that **MLflow** uses to record runs. If None, then it tries to get the environment variable **\'MLFLOW_PATH\'** or the **\'HOMEPATH\'** of the user. \n> \n> **name : string, optional (default=None)** <br> Name of the **MLflow** experiment to track.\n>\n> **mapping_paths : list, optional (default=None)** <br> Absolute paths to additional mapping files.\n>\n> **mapping : dict, optional (default=None)** <br> Mapping to the logging functions to use for the tracking of the events. If None, then a DEFAULT_MAPPING is used which allow to log parameters, outputs or inputs.\n>\n> **init_run_fns : list, optional (default=None)** <br> Logging function to execute on tracking initialization.\n>\n> **include_default_mappings : boolean, optional (default=True)** <br> A flag whether to use the default provided mappings or not.\n>\n> **logging_fns : dict, optional (default=None)** <br> User defined logging functions to use where each dict item has to be \' "event": fn\' or \' "event": {fn1,fn2,...}\'.\n>\n> **config : dict, optional (default=None)** <br> A dictionary that maps the events defined in PyPads mapping files with the logging functions.\n>\n> **reload_modules : boolean, optional (default=False)** <br> Reload and duck punch already loaded modules before the tracking activation if set to True.\n>\n> **clear_imports: boolean, optional (default=False)** <br> Delete alredy loaded modules for sys.modules() if set to True.\n# Scientific work disclaimer\nThis was created in scope of scientific work of the Data Science Chair at the University of Passau. If you want to use this tool or any of its resources in your scientific work include a citation.\n\n# Acknowledgement\nThis work has been partially funded by the **Bavarian Ministry of Economic Affairs, Regional Development and Energy** by means of the funding programm **"Internetkompetenzzentrum Ostbayern"** as well as by the **German Federal Ministry of Education and Research** in the project **"Provenance Analytics"** with grant agreement number *03PSIPT5C*.\n',
    'author': 'Thomas Weißgerber',
    'author_email': 'thomas.weissgerber@uni-passau.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.padre-lab.eu/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



import inspect
from _py_abc import ABCMeta
from abc import abstractmethod
from copy import copy

from pypads import logger

DEFAULT_ORDER = 1


class Context:
    __metaclass__ = ABCMeta

    def __init__(self, context):
        if context is None:
            raise ValueError("A context has to be passed for a object to be wrapped.")
        self._c = context

    def overwrite(self, key, obj):
        setattr(self._c, key, obj)

    def has_wrap_meta(self, mapping, wrappee):
        if not inspect.isfunction(wrappee):
            holder = wrappee
        else:
            holder = self._c
        # for k in holder.__dict__:
        #     if k.startswith("_pypads_mapping_"):
        #         if mapping in getattr(holder, k):
        #             return True
        if hasattr(holder, "_pypads_mapping_" + wrappee.__name__):
            if mapping in getattr(holder, "_pypads_mapping_" + wrappee.__name__):
                return True
        return False

    def store_wrap_meta(self, mapping, wrappee):
        try:
            if not inspect.isfunction(wrappee):
                holder = wrappee
            else:
                holder = self._c
            # Set self reference
            if not hasattr(holder, "_pypads_mapping_" + wrappee.__name__):
                setattr(holder, "_pypads_mapping_" + wrappee.__name__, [])
            getattr(holder, "_pypads_mapping_" + wrappee.__name__).append(mapping)
        except TypeError as e:
            logger.debug("Can't set attribute '" + wrappee.__name__ + "' on '" + str(self._c) + "'.")
            raise e

    def store_original(self, wrappee):
        try:
            if not inspect.isfunction(wrappee):
                holder = wrappee
            else:
                holder = self._c
            setattr(holder, self.original_name(wrappee), copy(wrappee))
        except TypeError as e:
            logger.debug("Can't set attribute '" + wrappee.__name__ + "' on '" + str(self._c) + "'.")
            return self._c

    def has_original(self, wrappee):
        return hasattr(self._c, self.original_name(wrappee)) or hasattr(wrappee,
                                                                        self.original_name(wrappee))

    def defined_stored_original(self, wrappee):
        if not inspect.isfunction(wrappee):
            return self.original_name(wrappee) in wrappee.__dict__
        else:
            return self.original_name(wrappee) in self._c.__dict__

    def original_name(self, wrappee):
        return "_pypads_original_" + str(wrappee.__name__)

    def original(self, wrappee):
        if not inspect.isfunction(wrappee) and not inspect.ismethod(wrappee):
            try:
                return getattr(wrappee, self.original_name(wrappee))
            except AttributeError:
                for attr in dir(wrappee):
                    if attr.endswith("_" + wrappee.__name__) and attr.startswith("_pypads_original_"):
                        return getattr(wrappee, attr)
        else:
            try:
                return getattr(self._c, self.original_name(wrappee))
            except AttributeError:
                for attr in dir(self._c):
                    if attr.endswith("_" + wrappee.__name__) and attr.startswith("_pypads_original_"):
                        return getattr(self._c, attr)

    def is_class(self):
        return inspect.isclass(self._c)

    def is_module(self):
        return inspect.ismodule(self._c)

    def real_context(self, fn_name):
        """
        Find where the function was defined
        :return:
        """

        # If the context is not an class it has to define the function itself
        if not self.is_class():
            if hasattr(self._c, fn_name):
                return self
            else:
                logger.warning("Context " + str(self._c) + " of type " + type(
                    self._c) + " doesn't define " + fn_name)
                return None

        # Find defining class by looking at the __dict__ and mro
        defining_class = None
        try:
            mro = self._c.mro()
            for clazz in mro[0:]:
                defining_class = clazz
                if hasattr(clazz, "__dict__") and fn_name in defining_class.__dict__:
                    if callable(defining_class.__dict__[fn_name]):
                        break
                    else:
                        # TODO do we need a workaround for
                        #  <sklearn.utils.metaestimators._IffHasAttrDescriptor object at 0x121e56810> again?
                        break
        except Exception as e:
            logger.warning("Couldn't get defining class of context '" + str(
                self._c) + ".")
            return self._c

        if defining_class and defining_class is not object:
            return Context(defining_class)
        return None

    @property
    def container(self):
        return self._c

    def get_dict(self):
        return self._c.__dict__

    def __str__(self):
        return str(self._c)


class BaseWrapper:
    __metaclass__ = ABCMeta

    def __init__(self, pypads):
        self._pypads = pypads

    @abstractmethod
    def wrap(self, wrappee, ctx, mapping):
        raise NotImplementedError()

    def _get_hooked_fns(self, fn, mapping):
        """
        For a given fn find the hook functions defined in a mapping and configured in a configuration.
        :param fn:
        :param mapping:
        :return:
        """
        if not mapping.hooks:
            mapping.hooks = mapping.in_collection.get_default_fn_hooks()

        library = None
        version = None
        if mapping.in_collection is not None:
            library = mapping.in_collection.lib
            version = mapping.in_collection.lib_version
        hook_events_of_mapping = [hook.event for hook in mapping.hooks if hook.is_applicable(mapping=mapping, fn=fn)]
        output = []
        config = self._get_current_config()
        events = [[k, v] for k, v in config["events"].items()]

        # sort event order
        events.sort(key=lambda e: e[1]["order"] if "order" in e[1] else DEFAULT_ORDER)

        for log_event, event_config in events:
            configured_hook_events = event_config["on"]

            # Add by config defined parameters
            if "with" in event_config:
                hook_params = event_config["with"]
            else:
                hook_params = {}

            # If one configured_hook_events is in this config.
            if configured_hook_events == "always" or set(configured_hook_events) & set(hook_events_of_mapping):
                fns = self._pypads.function_registry.find_functions(log_event, lib=library, version=version)
                if fns:
                    # sort logger order
                    fns = list(fns)
                    fns.sort(key=lambda f: f.order())
                    for fn in fns:
                        output.append((fn, hook_params))
        return output

    @classmethod
    def _get_current_config(cls):
        from pypads.pypads import get_current_config
        return get_current_config(default={"events": {}, "recursive": True})



from abc import ABCMeta, abstractmethod

from pypads import logger
from pypads.functions.loggers.base_logger import FunctionHolder
from pypads.functions.loggers.mixins import DependencyMixin, OrderMixin, IntermediateCallableMixin, TimedCallableMixin, \
    DefensiveCallableMixin


class PostRunFunction(DefensiveCallableMixin, IntermediateCallableMixin, FunctionHolder, TimedCallableMixin,
                      DependencyMixin, OrderMixin):
    """
    This class should be used to define new post run functions
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, fn=None, **kwargs):
        super().__init__(*args, fn=fn, **kwargs)
        if self._fn is None:
            self._fn = self._call

    @abstractmethod
    def _call(self, pads, *args, **kwargs):
        """
        Function where to add you custom code to execute after ending the run.

        :param pads: the current instance of PyPads.
        """
        return NotImplementedError()

    def __real_call__(self, *args, **kwargs):
        from pypads.pypads import get_current_pads
        return super().__real_call__(get_current_pads(), *args, **kwargs)

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        logger.warning(str(error))



from .schema import Attribute, Schema, ListAttribute, AlgorithmSchema


default_pipeline_schema = Schema(
    {
        "steps": ListAttribute("steps", "The steps", False, [
            {
                "docs": Attribute("docs", "Docstring", True, str),
                "algorithm": Attribute("algorithm", "The name of the used algorithm", False, str)
            },
            AlgorithmSchema()
        ]),
        "docs": Attribute("docs", "Docstring", True, str)
    }
)
